import unittest
from genomfart.utils.bigDataFrame import BigDataFrame
from genomfart.data.data_constants import FRAME_TEST_FILE
from Ranger import RangeSet, Range
import numpy as np

debug = False

class bigDataFrameTest(unittest.TestCase):
    """ Tests for bigDataFrame.py """
    @classmethod
    def setUpClass(cls):
        cls.frame = BigDataFrame(FRAME_TEST_FILE)
    def test_getitem(self):
        if debug: print("Testing __getitem__")
        self.assertEqual(self.frame[1,2],27140818)
        self.assertEqual(self.frame[1,'pos'],27140818)
        self.assertEqual(self.frame[6,'pos'],146678420)
        self.assertEqual(self.frame[0,'pos'],146650283)
        self.assertAlmostEqual(self.frame[1,'fval'],36.76025729538268)
        row = self.frame[1]
        self.assertAlmostEqual(row['fval'],36.76025729538268)
    def test_iter(self):
        if debug: print("Testing __iter__")
        row_iter = iter(self.frame)
        row = next(row_iter)
        self.assertEqual(row['pos'],146650283)
        row = next(row_iter)
        self.assertEqual(row['pos'],27140818)
        for row in row_iter:
            try:
                pass
            except StopIteration:
                break
    def test_iterrow(self):
        if debug: print("Testing iterrows")
        # Regular row iterator
        row_iter = self.frame.iterrows(cache=False)
        row = next(row_iter)
        self.assertEqual(row['pos'],146650283)
        row = next(row_iter)
        self.assertEqual(row['pos'],27140818)
        for row in row_iter:
            try:
                pass
            except StopIteration:
                break
        # Column specific row iterator
        row_iter = self.frame.iterrows(colspec=('chr','pos','allele'))
        row = next(row_iter)
        self.assertEqual(len(row),3)
        self.assertEqual(row['pos'],146650283)
        self.assertEqual(row['chr'],10)
        self.assertEqual(row['allele'],'C/A_hm2')
        row = next(row_iter)
        self.assertEqual(len(row),3)
        self.assertEqual(row['pos'],27140818)
        self.assertEqual(row['chr'],10)
        self.assertEqual(row['allele'],'A/T_hm2')
    def test_make_numpy_array(self):
        if debug: print("Testing make_numpy_array")
        arr = self.frame.make_numpy_array(rows=[1,2,4,5],cols=['pos','cm'])
        self.assertEqual(arr.shape,(4,2))
        self.assertEqual(arr[0,0],27140818)
        self.assertAlmostEqual(arr[2,1],43.47540355575215)
        arr2 = self.frame.make_numpy_array(rows=[1,2,4,5],cols=[2,3])
        self.assertTrue(np.allclose(arr,arr2))
        arr2 = self.frame.make_numpy_array(rows=[1,2,4,5],cols=RangeSet(
            ranges=[Range.closed(2,3)]
        ))
        self.assertTrue(np.allclose(arr,arr2))
        arr2 = self.frame.make_numpy_array(rows=[1,2,4,5],cols=RangeSet(
            ranges=[Range.closedOpen(2,3), Range.closedOpen(3,4)]
        ))
        self.assertTrue(np.allclose(arr,arr2))
        arr2 = self.frame.make_numpy_array(rows=RangeSet(
            ranges = [Range.closed(1,2), Range.closed(4,5)]
        ),cols=[2,3])
        self.assertTrue(np.allclose(arr,arr2))
if __name__ == '__main__':
    debug = True
    unittest.main(exit = False)    



import unittest
from Ranger import Range
from genomfart.parsers.gff import gff_parser
from genomfart.data.data_constants import GFF_TEST_FILE

debug = False

class gff_parserTest(unittest.TestCase):
    """ Tests for gff.py """
    @classmethod
    def setUpClass(cls):
        cls.parser = gff_parser(GFF_TEST_FILE)
    def test_get_overlapping_element_ids(self):
        if debug: print("Testing get_overlapping_element_ids")
        elements = self.parser.get_overlapping_element_ids('Pt',100,4000)
        self.assertEqual(len(elements),12)
        self.assertTrue('repeat_region:Pt_320_1262:+' in elements)
        self.assertTrue('repeat_region:Pt_3550_3560:?' in elements)
        self.assertTrue('repeat_region:Pt_3683_3696:?' in elements)
        self.assertTrue('repeat_region:Pt_3764_3775:?' in elements)
        self.assertTrue('exon:Pt_1674_3308:-' in elements)
        self.assertTrue('gene:GRMZM5G836994' in elements)
        self.assertTrue('transcript:GRMZM5G836994_T01' in elements)
        self.assertFalse('CDS:GRMZM5G811749_P01' in elements)
        self.assertTrue('CDS:GRMZM5G836994_P01' in elements)
        # Test with no overlapping elements
        elements = self.parser.get_overlapping_element_ids('Pt',1300,1310)
        self.assertEqual(len(elements),0)
    def test_get_overlapping_element_ids_of_type(self):
        if debug: print("Testing get_overlapping_element_ids_of_type")
        elements = self.parser.get_overlapping_element_ids_of_type('Pt',100,4000,'repeat_region')
        self.assertEqual(len(elements), 4)
        self.assertTrue('repeat_region:Pt_320_1262:+' in elements)
        self.assertTrue('repeat_region:Pt_3550_3560:?' in elements)
        self.assertTrue('repeat_region:Pt_3683_3696:?' in elements)
        self.assertTrue('repeat_region:Pt_3764_3775:?' in elements)        
    def test_get_element_info(self):
        if debug: print("Testing get_element_info")
        gene_info = self.parser.get_element_info('gene:GRMZM5G811749')
        self.assertEqual(gene_info['type'],'gene')
        self.assertEqual(gene_info['seqid'], 'Pt')
        self.assertEqual(len(gene_info['Ranges']),1)
        self.assertEqual(gene_info['Ranges'][0],Range.closed(3363,5604))
        self.assertEqual(len(gene_info['attributes']),1)
        self.assertEqual(gene_info['attributes'][0]['external_name'],'RPS16')
        self.assertEqual(gene_info['attributes'][0]['biotype'],'protein_coding')
    def test_get_element_ids_of_type(self):
        if debug: print("Testing get_element_ids_of_type")
        iterator = self.parser.get_element_ids_of_type('Pt','gene',start=100,
                                                       end=4000)
        self.assertEqual(next(iterator), 'gene:GRMZM5G836994')
        self.assertEqual(next(iterator), 'gene:GRMZM5G811749')
        with self.assertRaises(StopIteration):
            next(iterator)
    def test_get_element_children_ids(self):
        if debug: print("Testing get_element_children_ids")
        children = self.parser.get_element_children_ids('gene:GRMZM5G811749')
        self.assertEqual(len(children),1)
        self.assertEqual(children[0],'transcript:GRMZM5G811749_T01')
        children = self.parser.get_element_children_ids('transcript:GRMZM5G811749_T01')
        self.assertEqual(len(children), 3)
        self.assertFalse('gene:GRMZM5G811749' in children)
        self.assertTrue('CDS:GRMZM5G811749_P01' in children)
    def test_get_element_parent_ids(self):
        if debug: print("Testing get_element_parent_ids")
        parents = self.parser.get_element_parent_ids('gene:GRMZM5G811749')
        self.assertEqual(len(parents),0)
        parents = self.parser.get_element_parent_ids('transcript:GRMZM5G811749_T01')
        self.assertEqual(len(parents),1)
        self.assertEqual(parents[0],'gene:GRMZM5G811749')
        parents = self.parser.get_element_parent_ids('CDS:GRMZM5G811749_P01')
        self.assertEqual(len(parents),1)
        self.assertEqual(parents[0],'transcript:GRMZM5G811749_T01')
    def test_get_closest_element_id(self):
        if debug: print("Testing get_closest_element_id")
        closest_ids = self.parser.get_closest_element_id('Pt',2000,2200)
        self.assertEqual(closest_ids, set(['gene:GRMZM5G836994','transcript:GRMZM5G836994_T01',
                                           'CDS:GRMZM5G836994_P01','exon:Pt_1674_3308:-']))
        closest_ids = self.parser.get_closest_element_id('Pt',15838,15840)
        self.assertEqual(closest_ids, set(['repeat_region:Pt_15787_15836:?']))
        closest_ids = self.parser.get_closest_element_id('Pt',15880,15882)
        self.assertEqual(closest_ids, set(['repeat_region:Pt_15885_15961:+']))
    def test_get_closest_element_id_of_type(self):
        if debug: print("Testing get_closest_element_id")
        closest_ids = self.parser.get_closest_element_id_of_type('Pt',2000,2200,'gene')
        self.assertEqual(closest_ids, set(['gene:GRMZM5G836994']))
        closest_ids = self.parser.get_closest_element_id_of_type('Pt',15838,15840,'repeat_region')
        self.assertEqual(closest_ids, set(['repeat_region:Pt_15787_15836:?']))
        closest_ids = self.parser.get_closest_element_id_of_type('Pt',15880,15882,'gene')
        self.assertEqual(closest_ids, set(['gene:GRMZM5G813608']))
    def test_get_cds_indices(self):
        if debug: print("Testing get_cds_indices")
        cds_inds = self.parser.get_cds_indices('Pt',3308)
        self.assertEquals(cds_inds['CDS:GRMZM5G836994_P01'],1)
        cds_inds = self.parser.get_cds_indices('Pt',21472)
        self.assertEquals(cds_inds['CDS:GRMZM5G877454_P01'],5)
        cds_inds = self.parser.get_cds_indices('Pt',4707)
        self.assertEqual(len(cds_inds),1)
        self.assertEquals(cds_inds['CDS:GRMZM5G811749_P01'],42)
    def test_get_aa_indices(self):
        if debug: print("Testing get_aa_indices")
        aa_inds = self.parser.get_aa_indices('Pt',3308)
        self.assertEquals(aa_inds['CDS:GRMZM5G836994_P01'],1)
        aa_inds = self.parser.get_aa_indices('Pt',21472)
        self.assertEquals(aa_inds['CDS:GRMZM5G877454_P01'],2)
        aa_inds = self.parser.get_aa_indices('Pt',4707)
        self.assertEqual(len(aa_inds),1)
        self.assertEquals(aa_inds['CDS:GRMZM5G811749_P01'],14)
    def test_get_codon_position(self):
        if debug: print("Testing get_codon_position")
        pos = self.parser.get_codon_position('Pt',3308)
        self.assertEquals(pos['CDS:GRMZM5G836994_P01'],1)
        pos = self.parser.get_codon_position('Pt',21472)
        self.assertEquals(pos['CDS:GRMZM5G877454_P01'],2)
        pos = self.parser.get_codon_position('Pt',4707)
        self.assertEqual(len(pos),1)
        self.assertEquals(pos['CDS:GRMZM5G811749_P01'],3)
if __name__ == "__main__":
    debug = True
    unittest.main(exit = False)



import re
import sys
import gzip
from Bio import pairwise2

if sys.version_info[0] > 2:
    xrange = range

## Define some regular expressions that can be used to recognize certain data types
# Integers
int_re = re.compile(r'^-*[\d]+$')
# Floats
float_re = re.compile(r'(^-*[\d]+\.[\d]+$|^-*[\d]{1}([\.]?[\d])*(E|e)-[\d]+$)')
# Multiple integers
multi_int_re = re.compile(r'^((-*[\d]+),*)+$')
# Multiple floats
multi_float_re = re.compile(r'^((-*[\d]+\.[\d]+$|^-*[\d]{1}([\.]?[\d])*(E|e)-[\d]),*)+$')
## Parser used for VCF (v4.1) files
class VCF_parser:
    """
    Parser for VCF files
    """
    def __init__(self, vcf_file):
        """
        Instantiates a parser for the VCF file

        Parameters
        ----------

        vcf_file : str
            The path to a VCF file (May or may not be gzipped)
        """
        self.version = None
        self.info_dict = {}
        self.header_dict = {}
        self.genotypes = set()
        self.current_line = 0
        self.start_genotype_byte = 0
        if vcf_file.endswith('.gz'):
            self.file_handle = gzip.open(vcf_file)
        else:
            self.file_handle = open(vcf_file)
        # Parse through the beginning of the file
        while 1:
            line = self.file_handle.readline()
            if line.startswith('##fileformat='):
                # Get VCF version
                self.version = line.split('=')[1]
            elif line.startswith('##INFO='):
                line = line.strip()
                # Get field information
                info = '='.join(line.split('=')[1:])[1:-1]
                non_quoted_fields = list(map(lambda x: x.split('='),\
                                        re.findall("(?<=,)[^\s=]+=[^\"\s,=]+",info)))
                quoted_fields = list(map(lambda x: x.split('='),\
                                    re.findall("(?<=,)[^\s=]+=\"[^\"]+\"",info)))
                info_id = re.search("(?<=ID=)[^\s,]+",info).group()
                self.info_dict[info_id] = dict((k,v) for k,v in list(non_quoted_fields+quoted_fields))
            elif line.startswith('#CHROM'):
                # Get the header and the genotypes available
                line = line[1:].strip().split('\t')
                self.header_dict = dict((k,i) for i,k in enumerate(line))
                self.genotypes = line[9:]
                # Set the starting genotype byte
                self.start_genotype_byte = self.file_handle.tell()
                # Break from the loop
                break
    def parse_geno_depths(self):
        """
        Iterates through the VCF file, getting the genotypes at each position

        Returns
        -------
        A generator that generates tuples of chrom,pos,(ref,alt1,alt2,...),
        {sample->(base_depths)}
        """
        # Go to the start of the genotyping part of the file
        file_seek_pos = self.file_handle.seek(self.start_genotype_byte,0)
        # Go through the file
        for line in self.file_handle:
            line = line.strip().split('\t')
            if len(line) < 2:
                continue
            # Get chromosome and position
            chrom,pos = map(int,line[:2])
            # Get the possible alleles
            alt_alleles = line[self.header_dict['ALT']]
            if alt_alleles != '.':
                alleles = tuple([line[self.header_dict['REF']]]+\
                            line[self.header_dict['ALT']].split(','))
            else:
                alleles = (line[self.header_dict['REF']],)
            # Get the index of the allele depth field
            AD_ind = line[self.header_dict['FORMAT']].split(':').index('AD')
            # Get the depths of each genotype
            geno_dict = {}
            for sample in self.genotypes:
                samp_line = line[self.header_dict[sample]]
                if samp_line.startswith('./.'): geno_dict[sample] = tuple(map(lambda x: 0,alleles))
                else:
                    samp_line = samp_line.split(':')
                    geno_dict[sample] = tuple(map(int,samp_line[AD_ind].split(',')[:len(alleles)]))
            yield chrom,pos,alleles,geno_dict
            self.current_line += 1
    def parse_select_geno_depths(self, genos, info_dict = False, use_chrom = None,
                                 start = None, end = None):
        """
        Iterates through the VCF file, getting selected genotypes at each position.
        Note thtat this assumes samples contain depths

        Parameters
        ----------
        genos : list
           The names of the genotypes you want
        info_dict : boolean
            Whether you want the info dict on the end of the return
        use_chrom : str, optional
            Optional chromosome on which to start the scan
        start : int, optional
            Optional place to start the scan. If chrom also specified, it will be
            within the chromsoome. Otherwise, it will be within the first chromosome
            to have at least the start point
        end : int, optional
            Optional place to end the scan. (Nested within chrom if specified)
        
        Returns
        -------
        A generator that generates tuples of chrom,pos,(ref,alt1,alt2,...),{sample->base_depths),
                                            <info_dict if desired>}
        """
        # Go to the start of the genotyping part of the file
        file_seek_pos = self.file_handle.seek(self.start_genotype_byte,0)
        # Go through the file
        for line in self.file_handle:
            line = line.strip().split('\t')
            if len(line) < 2:
                continue
            # Get chromosome and position
            chrom,pos = line[0],int(line[1])
            if (use_chrom is not None and chrom != use_chrom):
                continue
            elif start and not end:
                if pos < start: continue
            elif end and not start:
                if pos > end: continue
            elif start and end:
                if not start <= pos <= end:
                    continue
            # Get the possible alleles
            alt_alleles = line[self.header_dict['ALT']]
            if alt_alleles != '.':
                alleles = tuple([line[self.header_dict['REF']]]+\
                            line[self.header_dict['ALT']].split(','))
            else:
                alleles = (line[self.header_dict['REF']],)
            # Get the index of the allele depth field
            AD_ind = line[self.header_dict['FORMAT']].split(':').index('AD')
            # Get the depths of each genotype
            geno_dict = {}
            for sample in genos:
                samp_line = line[self.header_dict[sample]]
                if samp_line.startswith('./.'): geno_dict[sample] = tuple(map(lambda x: 0,alleles))
                else:
                    samp_line = samp_line.split(':')
                    geno_dict[sample] = tuple(map(int,samp_line[AD_ind].split(',')[:len(alleles)]))
            ## Create dictionary for info if necessary
            line_info_dict = {}
            if info_dict:
                for field in line[self.header_dict['INFO']].split(';'):
                    if '=' in field:
                        key,val = field.split('=')
                        # Check for regular expressions
                        if int_re.match(val):
                            val = int(val)
                        elif float_re.match(val):
                            val = float(val)
                        elif multi_int_re.match(val):
                            val = tuple(map(int, val.split(',')))
                        elif multi_float_re.match(val):
                            val = tuple((map(float,val.split(','))))
                        line_info_dict[key] = val
                    else:
                        key = field
                        line_info_dict[key] = None
            if info_dict:
                yield chrom,pos,alleles,geno_dict,line_info_dict
            else:
                yield chrom,pos,alleles,geno_dict
            self.current_line += 1
    def parse_select_geno_generic(self, genos, info_dict = False, use_chrom = None,
                                   start = None, end = None, filter_excludes = None,
                                   filter_requires = None):
        """
        Iterates through the VCF file, getting selected genotypes at each position.
        This makes no assumptions about the format of the sample information for each genotype

        Parameters
        ----------

        genos : list
            The names of the genotypes you want
        info_dict : dict
            Whether you want the info dict on the end of the return
        use_chrom : str, optional
            Optional chromosome on which to start the scan
        start : int, optional
            Optional place to start the scan
        end : int, optional
            Optional place to end the scan
        filter_excludes : set, optional
            Filter tags that should exclude the locus from being returned
        filter_requires : set, optional
            Filter tags that should be required for a locus to be returned
        
        Returns
        -------
        A generator that generates tuples of chrom,pos,(ref,alt1,alt2,...),{sample->{prefix->val}),
                                            <info_dict if desired>}
        """
        if filter_excludes is None:
            filter_excludes = frozenset()
        else:
            filter_excludes = frozenset(filter_excludes)
        if filter_requires is None:
            filter_requires = frozenset()
        else:
            filter_requires = frozenset(filter_requires)
        # Go to the start of the genotyping part of the file
        file_seek_pos = self.file_handle.seek(self.start_genotype_byte,0)
        # Go through the file
        for line in self.file_handle:
            line = line.strip().split('\t')
            if len(line) < 2:
                continue
            # Get filter tags and filter if necessary
            filt = line[self.header_dict['FILTER']].split(',')
            if len(filter_excludes.intersection(filt)) > 0: continue
            elif len(filter_requires.intersection(filt)) != len(filter_requires): continue            
            # Get chromosome and position
            chrom,pos = line[0],int(line[1])
            if (use_chrom is not None and chrom != use_chrom):
                continue
            elif start and not end:
                if pos < start: continue
            elif end and not start:
                if pos > end: continue
            elif start and end:
                if not start <= pos <= end:
                    continue
            # Get the possible alleles
            alt_alleles = line[self.header_dict['ALT']]
            if alt_alleles != '.':
                alleles = tuple([line[self.header_dict['REF']]]+\
                            line[self.header_dict['ALT']].split(','))
            else:
                alleles = (line[self.header_dict['REF']],)
            # Get info of each genotype
            geno_keys = line[self.header_dict['FORMAT']].split(':')
            geno_dict = {}
            for sample in genos:
                samp_line = line[self.header_dict[sample]].split(':')
                try:
                    geno_dict[sample] = dict((k,samp_line[i]) for i,k in enumerate(geno_keys))
                except IndexError:
                    geno_dict[sample] = {geno_keys[0]: samp_line[0]}
            ## Create dictionary for info if necessary
            line_info_dict = {}
            if info_dict:
                for field in line[self.header_dict['INFO']].split(';'):
                    if '=' in field:
                        key,val = field.split('=')
                        # Check for regular expressions
                        if int_re.match(val):
                            val = int(val)
                        elif float_re.match(val):
                            val = float(val)
                        elif multi_int_re.match(val):
                            val = tuple(map(int, val.split(',')))
                        elif multi_float_re.match(val):
                            val = tuple((map(float,val.split(','))))
                        line_info_dict[key] = val
                    else:
                        key = field
                        line_info_dict[key] = None
            if info_dict:
                yield chrom,pos,alleles,geno_dict,line_info_dict
            else:
                yield chrom,pos,alleles,geno_dict
            self.current_line += 1
    def parse_site_infos(self, filter_excludes = None, filter_requires = None):
        """
        Iterates through the VCF file, getting the info for each site

        Parameters
        ----------
        filter_excludes : set, optional
            Filter tags that should exclude the locus from being returned
        filter_requires : set, optional
            Filter tags that should be required for a locus to be returned

        Returns
        -------
        A generator that generates tuples of chrom,pos,{ref,alt1,alt2,...),{field->val}.
        Fields without a corresponding value will have value "None"
        """
        if filter_excludes is None:
            filter_excludes = set()
        if filter_requires is None:
            filter_requires = set()
        # Go to the start of the genotyping part of the file
        file_seek_pos = self.file_handle.seek(self.start_genotype_byte,0)
        # Go through the file
        for line in self.file_handle:
            line = line.lstrip().split('\t',self.header_dict['INFO']+2)
            if len(line) < 2: continue
            # Get chromosome and position
            chrom,pos = map(int,line[:2])
            # Get quality and filter if necessary
            filt = line[self.header_dict['FILTER']].split(',')
            if len(filter_excludes.intersection(filt)) > 0: continue
            elif len(filter_requires.intersection(filt)) != len(filter_requires): continue
            # Get the possible alleles
            alt_alleles = line[self.header_dict['ALT']]
            if alt_alleles != '.':
                alleles = tuple([line[self.header_dict['REF']]]+\
                                line[self.header_dict['ALT']].split(','))
            else:
                alleles = (line[self.header_dict['REF']],)
            ## Create dictionary for info
            info_dict = {}
            for field in line[self.header_dict['INFO']].split(';'):
                if '=' in field:
                    key,val = field.split('=')
                    # Check for regular expressions
                    if int_re.match(val):
                        val = int(val)
                    elif float_re.match(val):
                        val = float(val)
                    elif multi_int_re.match(val):
                        val = tuple(map(int, val.split(',')))
                    elif multi_float_re.match(val):
                        val = tuple((map(float,val.split(','))))
                    info_dict[key] = val
                else:
                    key = field
                    info_dict[key] = None
            yield chrom, pos, alleles, info_dict
            self.current_line += 1
    @staticmethod
    def get_affected_ref_bases(vcf_pos, ref_allele, alt_allele):
        """ Gets the reference positions that are modified through
        the alternative allele, either through substitution or deletion.
        Note that this assumes that the first bases of the ref and alt
        alleles line up and that there is, at most, one indel in the
        alt_allele. 
        
        Parameters
        ----------
        vcf_pos : int
            The position given for this variant in the VCF file
        ref_allele : str
            The reference allele given by the VCF file
        alt_allele : str
            The alternative allele given by the VCF file

        Returns
        -------
        Set of positions that are either substituted or deleted in the
        alternative alleles

        Examples
        --------
        >>> VCF_parser.get_affected_ref_bases(20,'C','T')
        set([20])
        >>> VCF_parser.get_affected_ref_bases(20,'C','CTAG')
        set([])
        >>> VCF_parser.get_affected_ref_bases(20,'TCG','T')
        set([21, 22])
        >>> VCF_parser.get_affected_ref_bases(20,'TCGCG','TCG')
        set([24, 23])
        >>> VCF_parser.get_affected_ref_bases(20,'TCGCG','TCGCGCG')
        set([])
        """
        bases = frozenset(['A','C','G','T'])
        affected_set = set()
        if len(ref_allele) == 1 and len(alt_allele) == 1:
            if alt_allele not in bases:
                raise TypeError("%s is not a valid allele" % alt_allele)
            elif ref_allele != alt_allele:
                return set([vcf_pos])
            else:
                return set()
        elif len(ref_allele) >= len(alt_allele):
            for i in xrange(len(ref_allele)):
                if i < len(alt_allele):
                    # Adding substituted bases
                    if ref_allele[i] != alt_allele[i]:
                        affected_set.add(i+vcf_pos)
                else:
                    # Adding deleted bases
                    affected_set.add(i+vcf_pos)
        elif len(ref_allele) < len(alt_allele):
            for i in xrange(len(ref_allele)):
                # Adding substituted bases
                if ref_allele[i] != alt_allele[i]:
                    affected_set.add(i+vcf_pos)
        return affected_set
    @staticmethod
    def get_substituted_ref_bases(vcf_pos, ref_allele, alt_allele):
        """ Gets the reference positions that are modified through
        the alternative allele through substitution.
        
        Note that this assumes that the first bases of the ref and alt
        alleles line up and that there is, at most, one indel in the
        alt_allele. 
        
        Parameters
        ----------
        vcf_pos : int
            The position given for this variant in the VCF file
        ref_allele : str
            The reference allele given by the VCF file
        alt_allele : str
            The alternative allele given by the VCF file

        Returns
        -------
        Set of positions that are either substituted in the alternative allele

        Examples
        --------
        >>> VCF_parser.get_substituted_ref_bases(20,'C','T')
        set([20])
        >>> VCF_parser.get_substituted_ref_bases(20,'C','CTAG')
        set([])
        >>> VCF_parser.get_substituted_ref_bases(20,'TCG','T')
        set([])
        >>> VCF_parser.get_substituted_ref_bases(20,'TCGCG','TCG')
        set([])
        >>> VCF_parser.get_substituted_ref_bases(20,'TCGCG','TCGGCGCG')
        set([24, 23])
        >>> VCF_parser.get_substituted_ref_bases(20,'TCGCG','TCGCGGCG')
        set([])
        """
        bases = frozenset(['A','C','G','T'])
        affected_set = set()
        if len(ref_allele) == 1 and len(alt_allele) == 1:
            if alt_allele not in bases:
                raise TypeError("%s is not a valid allele" % alt_allele)
            elif ref_allele != alt_allele:
                return set([vcf_pos])
            else:
                return set()
        elif len(ref_allele) >= len(alt_allele):
            for i in xrange(len(alt_allele)):
                # Adding substituted bases
                if ref_allele[i] != alt_allele[i]:
                    affected_set.add(i+vcf_pos)
        elif len(ref_allele) < len(alt_allele):
            for i in xrange(len(ref_allele)):
                # Adding substituted bases
                if ref_allele[i] != alt_allele[i]:
                    affected_set.add(i+vcf_pos)
        return affected_set
    @staticmethod
    def get_nw_aligned_alleles(ref_allele, alt_allele, match = 1, mismatch = -2,
                               gapopen=-4, gapextend=-1):
        """ Aligns alleles using the Needleman-Wunsch global alignment algorithm

        Note that this will only return 1 of the alignments with the given score
        Parameters. This also assumes that the VCF always has the first bases of the
        allele aligned.
        ----------

        ref_allele : str
            The reference allele given by the VCF file
        alt_allele : str
            The alternative allele given by the VCF file
        match : number
            Score for matching a base
        mismatch : number
            Score for mismatching a base
        gapopen : number
            Score for opening a gap
        gapextend : number
            Score for extending a gap

        Returns
        -------
        aligned_seq1, aligned_seq2, score

        Examples
        --------
        >>> VCF_parser.get_nw_aligned_alleles('C','T')
        ('C', 'T', -2)
        >>> VCF_parser.get_nw_aligned_alleles('C','CTAG')
        ('C---', 'CTAG', -5)
        >>> VCF_parser.get_nw_aligned_alleles('TCG','T')
        ('TCG', 'T--', -4)
        >>> VCF_parser.get_nw_aligned_alleles('TCGCG','TCG')
        ('TCGCG', 'TC--G', -2.0)
        >>> VCF_parser.get_nw_aligned_alleles('TCGCG','TCGGCGCG')
        ('TCG---CG', 'TCGGCGCG', -1.0)
        >>> VCF_parser.get_nw_aligned_alleles('TCGCG','TCGCGGCG')
        ('TCGCG---', 'TCGCGGCG', -1.0)
        """
        aln_ref = None
        aln_alt = None
        score = 0.
        if len(ref_allele) == 1:
            aln_ref = ref_allele + '-'*(len(alt_allele)-len(ref_allele))
            aln_alt = alt_allele
            score = match if aln_ref[0] == aln_alt[0] else mismatch
            if len(alt_allele) > len(ref_allele):
                score += (gapopen + gapextend*(len(alt_allele)-len(ref_allele)-1))
        elif len(alt_allele) == 1:
            aln_ref = ref_allele
            aln_alt = alt_allele + '-'*(len(ref_allele)-len(alt_allele))
            score = match if aln_ref[0] == aln_alt[0] else mismatch
            if len(ref_allele) > len(alt_allele):
                score += (gapopen + gapextend*(len(ref_allele)-len(alt_allele)-1))
        else:
            alignments = pairwise2.align.globalms(ref_allele[1:], alt_allele[1:], match, mismatch,
                                                  gapopen, gapextend)
            aln_ref, aln_alt, score = alignments[0][0], alignments[0][1], alignments[0][2]
            aln_ref = ref_allele[0] + aln_ref
            aln_alt = alt_allele[0] + aln_alt
            score += (match if aln_ref[0] == aln_alt[0] else mismatch)
        return aln_ref, aln_alt, score
    @staticmethod
    def get_substituted_ref_bases_nw(vcf_pos, ref_allele, alt_allele, match = 1, mismatch = -2,
                               gapopen=-4, gapextend=-1):
        """ Gets the reference positions that are modified through
        the alternative allele through substitution. This is accomplished by
        first running a Needleman-Wunsch alignment of the 2 alleles and then
        finding the substitutions
        
        Note that this assumes that the first bases of the ref and alt
        alleles line up
        
        Parameters
        ----------
        vcf_pos : int
            The position given for this variant in the VCF file
        ref_allele : str
            The reference allele given by the VCF file
        alt_allele : str
            The alternative allele given by the VCF file
        match : number
            Score for matching a base
        mismatch : number
            Score for mismatching a base
        gapopen : number
            Score for opening a gap
        gapextend : number
            Score for extending a gap            

        Returns
        -------
        Set of positions that are either substituted in the alternative allele

        Examples
        --------
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'C','T')
        set([20])
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'C','CTAG')
        set([])
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'TCG','T')
        set([])
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'TCGCG','TCG')
        set([])
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'TCGCG','TCGGCGCG')
        set([])
        >>> VCF_parser.get_substituted_ref_bases_nw(20,'TCGCG','TCGCGGCG')
        set([])
        """
        bases = frozenset(['A','C','G','T'])
        affected_set = set()
        if len(ref_allele) == 1 and len(alt_allele) == 1:
            if alt_allele not in bases:
                raise TypeError("%s is not a valid allele" % alt_allele)
            elif ref_allele != alt_allele:
                return set([vcf_pos])
            else:
                return set()
        else:
            # Get NW alignment
            aln = VCF_parser.get_nw_aligned_alleles(ref_allele, alt_allele, match=match,
                                                    mismatch=mismatch, gapopen=gapopen,
                                                    gapextend=gapextend)
            add_num = 0
            for i, ref_base in enumerate(aln[0]):
                if ref_base == '-':
                    add_num += 1
                    continue
                elif aln[1][i] == '-':
                    continue
                elif ref_base != aln[1][i]:
                    affected_set.add(vcf_pos+i+add_num)
        return affected_set            
        



from bisect import bisect_right, bisect_left
from Ranger import Range
from genomfart.utils.bigDataFrame import BigDataFrame
import networkx as nx
import sys

if sys.version_info[0] > 2:
    xrange = range

## Instantiated class used to map between two versions of an assembly
class version_mapper:
    ## Graph, where nodes are named as (v1/v2, chrom, end_pt_of_region).
    #regions are connected by edges
    graph = None
    ## Dictionary of chrom -> [end points of regions in v1 assembly]
    # Same order as v2_end_dict for matching region
    v1_end_dict = {}
    ## Dictionary of chrom -> [end points of regions in v2 assembly]
    # Same order as v1_end_dict for matching region
    v2_end_dict = {}
    ## Instantiates the version mapper
    # @param map_file A file that maps between the two assemblies. It should have
    # columns v1_chrom, v1_start, v1_end, v2_chrom, v2_start, v2_end, orientation.
    # It is also assumed to be base-1
    def __init__(self, map_file):
        """
        Instantiates the version mapper

        Parameters
        ----------
        map_file : str
            A file that maps between the two assemblies. It should have columns
            v1_chrom, v1_start, v1_end, v2_chrom, v2_start, v2_end, orientation.
            It is also assumed to be base-1
        """
        map_frame = BigDataFrame(map_file,header=True, assume_uniform_types=False)
        ## Insantiate the dictionaries
        self.graph = nx.Graph()
        self.v1_end_dict = {}
        self.v2_end_dict = {}
        ## Fill the dictionaries
        for i,row in enumerate(map_frame):
            # Add chromosomes to endpoint dictionaries if necessary
            if row['v1_chrom'] not in self.v1_end_dict:
                self.v1_end_dict[row['v1_chrom']] = []
            if row['v2_chrom'] not in self.v2_end_dict:
                self.v2_end_dict[row['v2_chrom']] = []
            # Put in start and end points
            self.v1_end_dict[row['v1_chrom']].append(row['v1_end'])
            self.v2_end_dict[row['v2_chrom']].append(row['v2_end'])
            # Put in edges in graph
            node1 = ('v1',row['v1_chrom'],row['v1_end'])
            node2 = ('v2',row['v2_chrom'],row['v2_end'])
            self.graph.add_edge(node1,node2, strand=row['orientation'])
            self.graph.node[node1]['start'] = row['v1_start']
            self.graph.node[node1]['end'] = row['v1_end']
            self.graph.node[node1]['chrom'] = row['v1_chrom']
            self.graph.node[node2]['start'] = row['v2_start']
            self.graph.node[node2]['end'] = row['v2_end']
            self.graph.node[node2]['chrom'] = row['v2_chrom']
        ## Sort chromosome lists
        for dic in (self.v1_end_dict,self.v2_end_dict):
            for k in dic:
                dic[k] = sorted(dic[k])
    ## Gets the position in the version 2 genome if it exists (base 1 assumed)
    # @param v1_chr The version 1 chromosome
    # @param v1_pos The version 1 position
    # @return (chr,pos,orientation) if position exists in version 2 or None if not
    def v1_to_v2_map(self, v1_chr, v1_pos):
        """
        Gets the position of the version 2 genome if it exists (base 1 assumed)

        Parameters
        ----------
        v1_chr : hashable
            The version 1 chromosome
        v1_pos : int
            The version 1 position

        Returns
        -------
        (v2_chrom,v2_pos,orientation relative to v1) if the position exists in version 2
        or None if it doesn't exist in version 2
        """
        # Get the index of the end position within v1 in the end dict
        v1_end_ind = bisect_left(self.v1_end_dict[v1_chr],v1_pos)
        if v1_end_ind >= len(self.v1_end_dict[v1_chr]): return None
        # Define the region node for v1
        v1_node = ('v1',v1_chr,self.v1_end_dict[v1_chr][v1_end_ind])
        v2_node = self.graph.neighbors(v1_node)[0]
        # Check if the position is actually within the v1 interval
        # If not, return None
        if not self.graph.node[v1_node]['start'] <= v1_pos <= self.graph.node[v1_node]['end']:
            return None
        else:
            orientation = self.graph.edge[v1_node][v2_node]['strand']
            spacing = v1_pos - self.graph.node[v1_node]['start']
            if orientation == 1:
                return (v2_node[1], self.graph.node[v2_node]['start'] + spacing,orientation)
            else:
                return (v2_node[1], self.graph.node[v2_node]['end'] - spacing,orientation)
    ## Gets the position in the version 1 genome if it exists (base 1 assumed)
    # @param v2_chr The version 2 chromosome
    # @param v2_pos The version 2 position
    # @return (chr,pos,orientation) if position exists in version 2 or None if not
    def v2_to_v1_map(self, v2_chr, v2_pos):
        """
        Gets the position of the version 1 genome if it exists (base 1 assumed)

        Parameters
        ----------
        v2_chr : hashable
            The version 2 chromosome
        v2_pos : int
            The version 2 position

        Returns
        -------
        (v1_chrom,v1_pos,orientation relative to v2) if the position exists in version 1
        or None if it doesn't exist in version 1
        """
        # Get the index of the end position within v2 in the end dict
        v2_end_ind = bisect_left(self.v2_end_dict[v2_chr],v2_pos)
        if v2_end_ind >= len(self.v2_end_dict[v2_chr]): return None
        # Define the region node for v2
        v2_node = ('v2',v2_chr,self.v2_end_dict[v2_chr][v2_end_ind])
        v1_node = self.graph.neighbors(v2_node)[0]
        # Check if the position is actually within the v2 interval
        # If not, return None
        if not self.graph.node[v2_node]['start'] <= v2_pos <= self.graph.node[v2_node]['end']:
            return None
        else:
            orientation = self.graph.edge[v2_node][v1_node]['strand']
            spacing = v2_pos - self.graph.node[v2_node]['start']
            if orientation == 1:
                return (v1_node[1], self.graph.node[v1_node]['start'] + spacing,orientation)
            else:
                return (v1_node[1], self.graph.node[v1_node]['end'] - spacing,orientation)
    def v1_to_v2_seg_map(self, v1_chr, v1_start, v1_end):
        """
        Gets the ranges of positions in the version 2 genome if they exist (base 1 assumed)

        Parameters
        ----------
        v1_chr : hashable
            The version 1 chromosome
        v1_start : int
            The version 1 start position (inclusive)
        v1_end : int
            The version 1 end position (inclusive)

        Returns
        -------
        Dictionary of (v1_chrom,v1_start,v1_end)->(v2_chrom,v2_start,v2_end,orientatoin relative to v1) for
        ranges of positions existing in version 2 
        """
        # Get index of version 1 segment corresponding to end position in the end dict
        v1_end_ind = min(bisect_left(self.v1_end_dict[v1_chr],v1_end),len(self.v1_end_dict[v1_chr])-1)
        # Get index of version 1 segment corresponding to start position in the end dict
        v1_start_ind = min(bisect_left(self.v1_end_dict[v1_chr],v1_start),len(self.v1_end_dict[v1_chr])-1)
        return_dict = {}
        # Go through the range of v1 nodes
        for i in xrange(min(v1_start_ind,v1_end_ind),max(v1_start_ind,v1_end_ind)+1):
            # Define the nodes
            v1_node = ('v1',v1_chr,self.v1_end_dict[v1_chr][i])
            v2_node = self.graph.neighbors(v1_node)[0]
            # Check if there is actual overlap between v1 interval and the start+end points
            ovlap = Range.closed(v1_start,v1_end).isConnected(Range.closed(self.graph.node[v1_node]['start'], \
                                                        self.graph.node[v1_node]['end']))
            
            if ovlap:
                orientation = self.graph.edge[v1_node][v2_node]['strand']
                spacing = max(v1_start,self.graph.node[v1_node]['start']) - \
                  self.graph.node[v1_node]['start']
                if orientation == 1:
                    key = (v1_chr,self.graph.node[v1_node]['start']+spacing,min(v1_end,\
                                                self.graph.node[v1_node]['end']))
                    seg_length = key[2]-key[1]
                    val = (v2_node[1],self.graph.node[v2_node]['start']+spacing,\
                           self.graph.node[v2_node]['start']+spacing+seg_length,orientation)
                    return_dict[key] = val
                else:
                    key = (v1_chr,self.graph.node[v1_node]['start']+spacing,min(v1_end,\
                                                self.graph.node[v1_node]['end']))
                    seg_length = key[2]-key[1]
                    val = (v2_node[1],self.graph.node[v2_node]['end']-spacing, \
                           self.graph.node[v2_node]['end']-spacing-seg_length,orientation)
                    return_dict[key] = val
        return return_dict
    def v2_to_v1_seg_map(self, v2_chr, v2_start, v2_end):
        """
        Gets the ranges of positions in the version 1 genome if they exist (base 1 assumed)

        Parameters
        ----------
        v2_chr : hashable
            The version 2 chromosome
        v2_start : int
            The version 2 start position (inclusive)
        v2_end : int
            The version 2 end position (inclusive)

        Returns
        -------
        Dictionary of (v2_chrom,v2_start,v2_end)->(v1_chrom,v1_start,v1_end,orientation relative to v2) for
        ranges of positions existing in version 1 
        """
        # Get index of version 2 segment corresponding to end position in the end dict
        v2_end_ind = min(bisect_left(self.v2_end_dict[v2_chr],v2_end),len(self.v2_end_dict[v2_chr])-1)
        # Get index of version 2 segment corresponding to start position in the end dict
        v2_start_ind = min(bisect_left(self.v2_end_dict[v2_chr],v2_start),len(self.v2_end_dict[v2_chr])-1)
        return_dict = {}
        # Go through the range of v2 nodes
        for i in xrange(min(v2_start_ind,v2_end_ind),max(v2_start_ind,v2_end_ind)+1):
            # Define the nodes
            v2_node = ('v2',v2_chr,self.v2_end_dict[v2_chr][i])
            v1_node = self.graph.neighbors(v2_node)[0]
            # Check if there is actual overlap between v2 interval and the start+end points
            ovlap = Range.closed(v2_start,v2_end).isConnected(Range.closed(self.graph.node[v2_node]['start'], \
                                                        self.graph.node[v2_node]['end']))
            if ovlap:
                orientation = self.graph.edge[v2_node][v1_node]['strand']
                spacing = max(v2_start,self.graph.node[v2_node]['start']) - \
                  self.graph.node[v2_node]['start']
                if orientation == 1:
                    key = (v2_chr,self.graph.node[v2_node]['start']+spacing,min(v2_end,\
                                                self.graph.node[v2_node]['end']))
                    seg_length = key[2]-key[1]
                    val = (v1_node[1],self.graph.node[v1_node]['start']+spacing,\
                           self.graph.node[v1_node]['start']+spacing+seg_length,orientation)
                    return_dict[key] = val
                else:
                    key = (v2_chr,self.graph.node[v2_node]['start']+spacing,min(v2_end,\
                                                self.graph.node[v2_node]['end']))
                    seg_length = key[2]-key[1]
                    val = (v1_node[1],self.graph.node[v1_node]['end']-spacing, \
                           self.graph.node[v1_node]['end']-spacing-seg_length,orientation)
                    return_dict[key] = val
        return return_dict
        
